"""
Module to define style classes.
"""

# pylint: disable=too-few-public-methods

from typing import TypedDict
from typing_extensions import Unpack
import osvg.float


def verify_hex_string(hex_value: str) -> bool:
    """
    Verify if given string represents a hexadecimal value
    """
    try:
        int(hex_value, 16)
    except (TypeError, ValueError):
        return False
    return True


def verify_color_value(value):
    """
    Verify if given string is a valid value for a color style attribute
    """
    if (
        value is not None
        and value != "none"
        and (
            not isinstance(value, str)
            or len(value) != 6
            or not verify_hex_string(value)
        )
    ):
        raise TypeError("hex_def must be a string of 6 hex values or 'none'")


class StyleStringAttribute:
    """
    Parent class for all style attributes.
    """

    key = ""

    def __init__(self, value: str = None) -> None:
        self._value = None
        self.value = value

    @property
    def value(self):
        """
        Get value
        """
        return self._value

    def _verify_value(self, value):
        """
        Verify value
        """
        if value is not None and not isinstance(value, str):
            raise ValueError("Style attribute value must be a string")

    @property
    def isdefault(self) -> bool:
        """
        Check, if object's value is not set.
        """
        return self.value is None

    def equals(self, other) -> bool:
        """
        Check, if other's string value is equal to self's string value
        """
        return self.value == other.value

    @value.setter
    def value(self, value):
        """
        Verify and set value
        """
        self._verify_value(value)
        self._value = value

    def __str__(self) -> str:
        return "" if self.value is None else f"{self.key}:{self.value}"

    def copy(self):
        """
        Return an object with the same string as value
        """
        return self.__class__(value=self.value)


class StyleFloatAttribute:
    """
    Parent class for all style attributes.
    """

    key = ""
    value = osvg.float.FloatProperty()
    str_round = 3

    def __init__(self, value: float | osvg.float.Float = None) -> None:
        self.value = value

    def __str__(self) -> str:
        return (
            ""
            if self.value.float_ref is None
            else f"{self.key}:{round(float(self.value), self.str_round):g}"
        )

    def copy(self):
        """
        Return an object with the same float value as value
        """
        return self.__class__(value=self.value.float_ref)

    @property
    def isdefault(self) -> bool:
        """
        Check, if object's value is not set.
        """
        return self.value.float_ref is None

    def equals(self, other) -> bool:
        """
        Check, if other's float value is equal to self's float value
        """
        if not self.isdefault and not other.isdefault:
            return float(self.value) == float(other.value)
        return self.isdefault == other.isdefault


class Color(StyleStringAttribute):
    """
    SVG representation of a color in string of hex values or "none".
    "rrggbb" ->:
        - "rr" 256 bit value for red
        - "gg" 256 bit value for green
        - "bb" 256 bit value for blue
    """

    def __init__(self, value: str = None) -> None:
        super().__init__(value=value)

    def _verify_value(self, value):
        verify_color_value(value)

    def __str__(self) -> str:
        if self.value == "none":
            return f"{self.key}:none"
        return "" if self.value is None else f"{self.key}:#{self.value}"


class Opacity(StyleFloatAttribute):
    """
    SVG representation of a opacity in float 0 <= value <= 1
    with 1 is 100% opacity.
    """

    key = "opacity"

    def __init__(self, value: float | osvg.float.Float = None) -> None:
        super().__init__(value=value)

    @property
    def opacity(self) -> float:
        """
        Get opacity float value
        """
        return self.value

    @opacity.setter
    def opacity(self, value: float | osvg.float.Float):
        """
        Set opacity
        """
        self.value = value


class FillColor(Color):
    """
    SVG Style Attribute "fill"
    """

    key = "fill"


class FillOpacity(Opacity):
    """
    SVG Style Attribute "fill-opacity"
    """

    key = "fill-opacity"


class StrokeColor(Color):
    """
    SVG Style Attribute "stroke"
    """

    key = "stroke"


class StrokeOpacity(Opacity):
    """
    SVG Style Attribute "stroke-opacity"
    """

    key = "stroke-opacity"


class StrokeWidth(StyleFloatAttribute):
    """
    SVG Style Attribute "stroke-width"
    """

    key = "stroke-width"

    def __init__(self, value: float | osvg.float.Float = None) -> None:
        super().__init__(value=value)


class StrokeLineCap(StyleStringAttribute):
    """
    SVG Style Attribute "stroke-width"
    """

    key = "stroke-linecap"

    def __init__(self, value: str = None) -> None:
        if value and value not in ["butt", "round", "square"]:
            raise ValueError(f"'{value}' not a valid linecap option")
        super().__init__(value=value)


class StrokeDashArray(osvg.float.FloatList):
    """
    SVG Style Attribute "stroke-width"
    """

    key = "stroke-dasharray"
    name = "values"

    def __str__(self) -> str:
        if not self.values:
            return ""
        return f"{self.key}:{super().__str__()}"


class StrokeLineJoin(StyleStringAttribute):
    """
    SVG Style Attribute "stroke-linejoin"
    """

    key = "stroke-linejoin"

    def __init__(self, value: str = None) -> None:
        if value and value not in [
            "arcs",
            "bevel",
            "miter",
            "miter-clip",
            "round",
        ]:
            raise ValueError(f"'{value}' not a valid linecap option")
        super().__init__(value=value)


class FontFamily(StyleStringAttribute):
    """
    SVG Font Style Attribute "font-family"
    """

    key = "font-family"


class FontSize(StyleFloatAttribute):
    """
    SVG Font Style Attribute "font-size" in pixels
    """

    key = "font-size"

    def __str__(self) -> str:
        return (
            ""
            if self.value.float_ref is None
            else f"{self.key}:{round(float(self.value), self.str_round):g}px"
        )


class FontSizeAdjust(StyleFloatAttribute):
    """
    SVG Font Style Attribute "font-size-adjust" in pixels
    """

    key = "font-size-adjust"


class FontStretch(StyleStringAttribute):
    """
    SVG Font Style Attribute "font-stretch"
    """

    key = "font-stretch"


class FontStyle(StyleStringAttribute):
    """
    SVG Font Style Attribute "font-style"
    """

    key = "font-style"
    allowed_values = [
        "normal",
        "italic",
        "oblique",
    ]

    def __init__(self, value: str = None) -> None:
        super().__init__(value)
        if value is not None and value not in self.allowed_values:
            raise ValueError(
                f"font-style value mus be on of {','.join(self.allowed_values)}"
            )


class FontVariant(StyleStringAttribute):
    """
    SVG Font Style Attribute "font-variant"
    """

    key = "font-variant"


class FontWeigth(StyleStringAttribute):
    """
    SVG Font Style Attribute "font-weigth"
    """

    key = "font-weight"
    allowed_values = [
        "normal",
        "bold",
        "bolder",
        "lighter",
    ]

    def __init__(self, value: str = None) -> None:
        super().__init__(value)
        if value is not None and value not in self.allowed_values:
            raise ValueError(
                f"font-weigth value must be on of {','.join(self.allowed_values)}"
            )


class StyleParamTypes(TypedDict):
    """
    Parameter type hint for class Style
    """

    opacity: Opacity | float | int
    fill_color: FillColor | str
    fill_opacity: FillOpacity | float | int
    stroke_color: StrokeColor | str
    stroke_opacity: StrokeOpacity | float | int
    stroke_width: StrokeWidth | float | int
    stroke_linecap: StrokeLineCap | str
    stroke_linejoin: StrokeLineJoin | str
    stroke_dasharray: StrokeDashArray | str
    font_family: FontFamily | str
    font_size: FontSize | float | int
    font_size_adjust: FontSizeAdjust | float | int
    font_stretch: FontStretch | str
    font_style: FontStyle | str
    font_variant: FontVariant | str
    font_weigth: FontWeigth | str
    reference: "Style"


class Style:
    """
    SVG Style representative.
    """

    param_classes = {
        "opacity": Opacity,
        "fill_color": FillColor,
        "fill_opacity": FillOpacity,
        "stroke_color": StrokeColor,
        "stroke_opacity": StrokeOpacity,
        "stroke_width": StrokeWidth,
        "stroke_linecap": StrokeLineCap,
        "stroke_linejoin": StrokeLineJoin,
        "stroke_dasharray": StrokeDashArray,
        "font_family": FontFamily,
        "font_size": FontSize,
        "font_size_adjust": FontSizeAdjust,
        "font_stretch": FontStretch,
        "font_style": FontStyle,
        "font_variant": FontVariant,
        "weigth": FontWeigth,
    }

    def __init__(self, **kwargs: Unpack[StyleParamTypes]) -> None:
        self.attributes = self.param_classes.copy()
        if "reference" in kwargs:
            ref_obj = kwargs["reference"]
            for attr in self.attributes:
                self.attributes[attr] = ref_obj.attributes[attr]
        for attr, _class in self.param_classes.items():
            if attr in kwargs:
                attr_obj = kwargs[attr]
                if not isinstance(attr_obj, _class):
                    attr_obj = _class(attr_obj)
                self.attributes[attr] = attr_obj
            elif "reference" not in kwargs:
                self.attributes[attr] = _class()

    def __str__(self) -> str:
        """
        SVG font style string
        """
        return_kvs = []
        for attr_v in self.attributes.values():
            if str(attr_v):
                return_kvs.append(str(attr_v))
        return ";".join(return_kvs)

    def copy(self) -> "Style":
        """
        Return an object with same atrributes.
        """
        obj = Style()
        for k, v in self.attributes.items():
            obj.attributes[k] = v.copy()
        return obj

    def merge(self, other: "Style") -> None:
        """
        Overwrite with all not 'None' values
        """
        for k, v in other.attributes.items():
            if isinstance(v, StyleStringAttribute):
                if v.value is not None:
                    self.attributes[k] = v.copy()
            if isinstance(v, StyleFloatAttribute):
                if v.value.float_ref is not None:
                    self.attributes[k] = v.copy()
            elif isinstance(v, StrokeDashArray):
                if v.values != []:
                    self.attributes[k] = StrokeDashArray(v.values)

    def diff(self, other: "Style") -> "Style":
        """
        Return style object with only attributes which are overwritten in 'other'
        """
        style = Style()
        for k, v in other.attributes.items():
            if isinstance(v, StyleStringAttribute):
                if v.value is not None and v.value != self.attributes[k].value:
                    style.attributes[k] = v.copy()
            if isinstance(v, StyleFloatAttribute):
                if not v.equals(self.attributes[k]) and not v.isdefault:
                    style.attributes[k] = v.copy()
            elif isinstance(v, StrokeDashArray):
                if v.values not in ([], self.attributes[k].values):
                    style.attributes[k] = StrokeDashArray(v.values)
        return style

    @property
    def stroke_width(self) -> float | None:
        """
        Get stroke width value
        """
        return (
            None
            if self.attributes[  # pylint: disable=using-constant-test
                "stroke_width"
            ].isdefault
            else float(self.attributes["stroke_width"].value)
        )



"""
Module to define Float classes to calculate float numbers.
"""

from collections.abc import Iterable
from types import GeneratorType
import math
import osvg.float
import osvg.helper


class EmptyCalculatedListError(ValueError):
    """
    Exception indicating an access to the value of CalculatedList object,
    which has no float values added yet.
    """


class Cos(osvg.float.OneValueMath):
    """
    Cosinus value of an angle in degree
    """

    def fill_cache(self):
        self._value_cache = math.cos(osvg.helper.radians(float(self.float_ref)))


class Sin(osvg.float.OneValueMath):
    """
    Sinus value of an angle in degree
    """

    def fill_cache(self):
        self._value_cache = math.sin(osvg.helper.radians(float(self.float_ref)))


class Cot(osvg.float.OneValueMath):
    """
    Cotangent value of an angle in degree
    """

    def fill_cache(self):
        v = osvg.helper.radians(float(self.float_ref))
        self._value_cache = math.cos(v) / math.sin(v)


class Tan(osvg.float.OneValueMath):
    """
    Tangent value of an angle in degree
    """

    def fill_cache(self):
        v = osvg.helper.radians(float(self.float_ref))
        self._value_cache = math.sin(v) / math.cos(v)


class CalculatedList(osvg.float.FloatMath):
    """
    Base class for all calculated Floats with a list of Float objects
    """

    def __init__(
        self,
        *args: Iterable[osvg.float.Float | float],
    ) -> None:
        super().__init__()
        if args:
            self.add(args)

    @property
    def floats(self) -> list[float]:
        """
        Get tuple of all float input object values
        """
        if not self.inputs:
            raise EmptyCalculatedListError(f"{self.__class__.__name__} has no inputs")
        return tuple(float(x) for x in self.inputs)

    def add(
        self, f: osvg.float.Float | float | Iterable[osvg.float.Float | float]
    ) -> None:
        """
        Add a Float value or a iterable or a generator object with Float values
        """
        if not isinstance(f, (Iterable, GeneratorType)):
            f = [f]
        for v in f:
            if not issubclass(type(v), osvg.float.Float):
                v = osvg.float.Float(v)
            self.add_input(v)

    def fill_cache(self) -> None:
        raise NotImplementedError


class CalculatedKwargs(osvg.float.FloatMath):
    """
    Base class for all calculated Floats with keyword Float objects arguments
    """

    def __init__(
        self,
        **kwargs: osvg.float.Float | float,
    ) -> None:
        super().__init__()
        self.__dict__.update(kwargs)
        for k, v in kwargs.items():
            if not isinstance(v, osvg.float.Float):
                v = osvg.float.Float(v)
            kwargs[k] = v
            self.add_input(v)
        self.__dict__.update(kwargs)

    def fill_cache(self) -> None:
        raise NotImplementedError


class FunctFloat(CalculatedList):
    """
    Class for customized calculation of Floats by passing a external function
    """

    def __init__(
        self,
        args: Iterable[osvg.float.Float | float | int],
        funct: "function",
    ) -> None:
        super().__init__(*args)
        self.funct = funct

    def fill_cache(self) -> None:
        args = self.floats
        self._value_cache = self.funct(*args)


class Min(CalculatedList):
    """
    Get the lowest value of all given Floats.
    """

    def fill_cache(self) -> None:
        """
        Lowest values of all given values
        """
        self._value_cache = min(self.floats)


class LowestAbs(CalculatedList):
    """
    Return the floating number with the lowest absolute value.
    """

    def fill_cache(self) -> None:
        """
        Lowest absolute value of all given values
        """
        floats = self.floats
        self._value_cache = floats[0]
        for v in floats[1:]:
            if abs(v) < abs(self._value_cache):
                self._value_cache = v


class Max(CalculatedList):
    """
    Get the highest value of all given Floats.
    """

    def fill_cache(self) -> None:
        """
        Highest values of all given values
        """
        self._value_cache = max(self.floats)


class PercentOf(CalculatedKwargs):
    """
    Calculate percentage of a value
    """

    def __init__(
        self,
        reference: osvg.float.Float,
        percentage: float | osvg.float.Float,
        modulo: bool = False,
    ) -> None:
        super().__init__(reference=reference, percentage=percentage)
        self._modulo = False
        self.modulo = modulo

    @property
    def modulo(self) -> bool:
        """
        Get modulo value
        """
        return self._modulo

    @modulo.setter
    def modulo(self, _bool: bool) -> None:
        self._modulo = _bool
        self.input_changed()

    def fill_cache(self) -> None:
        # pylint: disable=no-member
        if self.modulo:
            self._value_cache = (
                float(self.reference) / 100 * (float(self.percentage) % 100)
            )
        else:
            self._value_cache = float(self.reference) / 100 * float(self.percentage)


class Int(osvg.float.OneValueMath):
    """
    Represents the integer value of a Float reference
    """

    def fill_cache(self) -> None:
        """
        Calculate the value for the cache
        """
        self._value_cache = int(float(self.float_ref))


class Abs(osvg.float.OneValueMath):
    """
    Represents the absulote value of a Float reference
    """

    def fill_cache(self) -> None:
        """
        Calculate the value for the cache
        """
        self._value_cache = abs(float(self.float_ref))


class Average(CalculatedList):
    """
    Represents the Average value of all Float references
    """

    def fill_cache(self) -> None:
        """
        Calculate the value for the cache
        """
        floats = self.floats
        self._value_cache = sum(floats) / len(floats)



"""
Module to provide helper classes
"""

import math


class WeightedList:
    """
    A weighted list is list of elements with integer based weights.
    The default weight is 0. On iteration elements with higher weight
    will be processed first.
    Elements with same weight will be processed as they were
    appended, prepended or inserted.
    """

    def __init__(self) -> None:
        self._weighted_elements = {}
        self._index = 0
        self._list = []

    def _add_weight(self, weight: int) -> None:
        """
        Add a weight to the list.
        """
        if weight not in self._weighted_elements:
            self._weighted_elements[weight] = []

    @property
    def weights(self) -> list[int]:
        """
        Returns a sorted list of weights.
        """
        return sorted(list(self._weighted_elements), reverse=True)

    def append(self, new_element: any, weight: int = 0) -> None:
        """
        Add a new element to the end of the list.
        Optional the weight can be specified.
        """
        self._add_weight(weight)
        self._weighted_elements[weight].append(new_element)

    def insert(self, index: int, new_element: any, weight: int = 0) -> None:
        """
        Insert a new element at a specific position (index) of the list.
        Optional the weight can be specified.
        """
        self._add_weight(weight)
        self._weighted_elements[weight].insert(index, new_element)

    def prepend(self, new_element: any, weight: int = 0) -> None:
        """
        Add a new element to the beginning of the list.
        Optional the weight can be specified.
        """
        self.insert(0, new_element, weight)

    def __len__(self) -> int:
        """
        Returns the length of the list.
        """
        length = 0
        for l in self._weighted_elements.values():
            length += len(l)
        return length

    @property
    def length(self) -> int:
        """
        Length of the list.
        """
        return len(self)

    def length_w(self, weight: int) -> int:
        """
        Return the amount of elements for a given weight.
        """
        if weight not in self._weighted_elements:
            return 0
        return len(self._weighted_elements[weight])

    def list_w(self, weight: int) -> list:
        """
        Returns the list for a given weight.
        """
        if weight not in self._weighted_elements:
            raise IndexError(f"Weight {weight} not in this list")
        return self._weighted_elements[weight]

    @property
    def list(self) -> list:
        """
        Returns all elements according to the weight.
        """
        l = []
        for w in self.weights:
            l += self._weighted_elements[w]
        return l

    def __iter__(self):
        self._index = 0
        self._list = self.list
        return self

    def __next__(self):
        if self._index == len(self):
            raise StopIteration
        self._index += 1
        return self._list[self._index - 1]

    def extend(self, other: "WeightedList"):
        """
        Add all other's weighted element to self
        """
        # pylint: disable=protected-access
        if not isinstance(other, WeightedList):
            raise TypeError("Can only extend with another WeightedList")
        for w, elements in other._weighted_elements.items():
            if w in self._weighted_elements:
                self._weighted_elements[w].extend(elements)
            else:
                self._weighted_elements[w] = elements


def radians(angle: float) -> float:
    """
    Return the radiant of an angle in degree.
    Note: angle=0 ist 3 o'clock and it goes clockwise.
    """
    return -math.radians(angle)


def degrees(radiant: float) -> float:
    """
    Return the angle of a radiant float value.
    Note: angle=0 ist 3 o'clock and it goes clockwise.
    """
    return -math.degrees(radiant)


def x_position_on_circle(angle: float = 0, radius: float = 1) -> float:
    """
    Calculate the delta for x of a point on a circle with radius 'radius'
    and the angle 'anlge' in degrees.
    Note: angle=0 ist 3 o'clock and it goes clockwise.
    """
    # 1. Get the cosine of the radiant
    # 2. multiply by the radius of the circle
    # 3. Return float rounded to 3 decimals
    return round(math.cos(radians(angle)) * radius, 3)


def y_position_on_circle(angle: float = 0, radius: float = 1) -> float:
    """
    Calculate the delta for y of a point on a circle with radius 'radius'
    and the angle 'anlge' in degrees.
    Note: angle=0 ist twelve o'clock and it goes clockwise.
    """
    # 1. Get the sine of the radiant
    # 2. Invert (multiple by -1) to get the delta in a SVG oriented coordinational system
    # 3. multiply by the radius of the circle
    # 4. Return float rounded to 3 decimals
    return round(math.sin(radians(angle)) * -1 * radius, 3)


def angle_a_b(b: tuple[float, float], a: tuple[float, float] = None) -> float:
    """
    Angle from position 'a' (x, y) to position 'b' (x, y) in degrees.
    Rotation is clockwise with 0 == "12 o'clock"
    """
    if a is None:
        a = (0, 0)
    dx = b[0] - a[0]
    dy = a[1] - b[1]  # due to the inverted y-axis
    return degrees(math.atan2(dy, dx))


def radian_a_b(b: tuple[float, float], a: tuple[float, float] = None) -> float:
    """
    Angle from position 'a' (x, y) to position 'b' (x, y) in radians.
    """
    if a is None:
        a = (0, 0)
    dx = b[0] - a[0]
    dy = a[1] - b[1]  # due to the inverted y-axis
    return math.atan2(dy, dx) % (2 * math.pi)



"""
Module to define class for SVG ellipse elements.
"""

import osvg.elements.elementbase


class CircleBase(osvg.elements.elementbase.SVGElement):
    """
    Meta class for circle like elements.
    """

    def create_default_connectors(self) -> None:
        """
        Create Positions as connectors and add them.
        """
        self.add_connector_at_angle(angle=225, name="top-left", respect_rotation=True)
        self.add_connector_at_angle(angle=270, name="top-center", respect_rotation=True)
        self.add_connector_at_angle(angle=315, name="top-right", respect_rotation=True)
        self.add_connector_at_angle(
            angle=180, name="center-left", respect_rotation=True
        )
        self.add_connector(self.position, "center")
        self.add_connector_at_angle(angle=0, name="center-right", respect_rotation=True)
        self.add_connector_at_angle(
            angle=135, name="bottom-left", respect_rotation=True
        )
        self.add_connector_at_angle(
            angle=90, name="bottom-center", respect_rotation=True
        )
        self.add_connector_at_angle(
            angle=45, name="bottom-right", respect_rotation=True
        )

    def add_connector_at_angle(
        self,
        angle: float | osvg.float.Float,
        name: str,
        respect_rotation: bool = False,
    ):
        """
        Add a connector at element's border at the angle
        in relation to the center.
        """
        if not isinstance(angle, osvg.float.Float):
            angle = osvg.float.Float(angle)
        if respect_rotation:
            calc_angle = angle
        else:
            calc_angle = angle - self.rotation
        relative_shifted_position = self.angled_position(angle=calc_angle)
        if not respect_rotation:
            relative_shifted_position = osvg.positions.RotatedPosition(
                origin=relative_shifted_position,
                center=osvg.positions.Position(),
                angle=self.rotation,
            )

        calculated_position = self.position + relative_shifted_position
        return self.add_connector(
            position=calculated_position,
            name=name,
            respect_rotation=respect_rotation,
        )

    def angled_position(self, angle: osvg.float.Float) -> osvg.positions.Position:
        """
        Return position on border at given angle
        """
        raise NotImplementedError



"""
Module to define line start and end markers.
"""

from xml.etree import ElementTree as ET
from typing import TypedDict
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.style


class MarkerParams(TypedDict):
    """
    Keyword argument definition for CircleMarker class.
    """

    color: str = None


class Marker:
    """
    Base class for start and end markers
    """

    # pylint: disable=too-few-public-methods

    stroke_width_factor = 3

    def __init__(
        self,
        parent: "osvg.line.Line | osvg.polyline.Polyline",
        position: osvg.positions.Position,
        angle: float | osvg.float.Float,
        **kwargs,
    ) -> None:
        self.parent = parent
        self.position = position
        self.angle = angle
        color = kwargs.get("color", None)
        osvg.style.verify_color_value(color)
        self.color = color

    @property
    def length(self) -> osvg.float.Float:
        """
        Return length of marker. (To appropriately shorten line)
        """
        return osvg.float.Float(0)

    def _add_style(self, element: ET.Element) -> None:
        """
        Add style attributes
        """
        element.set("stroke-width", "0")
        self._add_color_style(element=element)

    def _add_color_style(self, element: ET.Element) -> None:
        """
        Add color style attribute
        """
        if self.color is not None:
            color = self.color
        else:
            color = self.parent.inherited_style.attributes["stroke_color"].value
        if color:
            element.set("fill", f"#{color}")

    @property
    def etree_element(self) -> ET.Element:
        """
        Get SVG element to draw marker
        """
        raise NotImplementedError()


class CircleMarkerParams(MarkerParams):
    """
    Keyword argument definition for CircleMarker class.
    """

    radius: float | osvg.float.Float = None


class CircleMarker(Marker):
    """
    Class to define a circle at the start/end of a line
    """

    def __init__(
        self,
        parent: "osvg.line.Line | osvg.polyline.Polyline",
        position: osvg.positions.Position,
        angle: float | osvg.float.Float,
        **kwargs: Unpack[CircleMarkerParams],
    ):
        super().__init__(parent=parent, position=position, angle=angle, **kwargs)
        self._radius = osvg.float.Float()
        self.radius = kwargs.get("radius", None)

    @property
    def radius(self) -> osvg.float.Float:
        """
        Get radius
        """
        return self._radius

    @radius.setter
    def radius(self, radius: float | osvg.float.Float | type[None]) -> None:
        if radius is None:
            parent_stroke_with = self.parent.style.attributes["stroke_width"]
            radius = (
                (
                    osvg.elements.elementbase.SVG_STROKE_WIDTH_DEFAULT
                    if parent_stroke_with.isdefault
                    else parent_stroke_with.value
                )
                * self.stroke_width_factor
                * 2
                / 3
            )
        if not isinstance(radius, osvg.float.Float):
            radius = osvg.float.Float(radius)
        self._radius = radius

    @property
    def etree_element(self) -> ET.Element:
        element = ET.Element("circle")
        element.set("cx", str(self.position.x))
        element.set("cy", str(self.position.y))
        element.set("r", str(self.radius))
        element.set("stroke-width", "0")
        self._add_style(element=element)
        return element


class ArrowMarkerParams(MarkerParams):
    """
    Keyword argument definition for ArrowMarker class.
    """

    length: float | osvg.float.Float = None
    width: float | osvg.float.Float = None


class ArrowMarker(Marker):
    """
    Class for regular Arrow Markers
    """

    angle = osvg.float.FloatProperty()

    def __init__(
        self,
        parent: "osvg.line.Line | osvg.polyline.Polyline",
        position: osvg.positions.Position,
        angle: float | osvg.float.Float,
        **kwargs: Unpack[ArrowMarkerParams],
    ):
        super().__init__(parent=parent, position=position, angle=angle, **kwargs)
        self._arrow_length = None
        self.arrow_length = kwargs.get("length", None)
        self._arrow_width = None
        self.arrow_width = kwargs.get("width", None)

    @property
    def length(self) -> osvg.float.Float:
        """
        Get length
        """
        return self.arrow_length

    @property
    def arrow_length(self) -> osvg.float.Float:
        """
        Get length
        """
        return self._arrow_length

    @arrow_length.setter
    def arrow_length(self, length: float | osvg.float.Float | type[None]) -> None:
        if length is None:
            parent_stroke_with = self.parent.style.attributes["stroke_width"]
            stroke_width = (
                osvg.elements.elementbase.SVG_STROKE_WIDTH_DEFAULT
                if parent_stroke_with.isdefault
                else parent_stroke_with.value
            )
            self._arrow_length = stroke_width * self.stroke_width_factor * 2
        else:
            if not isinstance(length, osvg.float.Float):
                length = osvg.float.Float(length)
            self._arrow_length = length

    @property
    def arrow_width(self) -> osvg.float.Float:
        """
        Get width
        """
        return self._arrow_width

    @arrow_width.setter
    def arrow_width(self, width: float | osvg.float.Float | type[None]) -> None:
        if width is None:
            self._arrow_width = self.arrow_length
        else:
            if not isinstance(width, osvg.float.Float):
                width = osvg.float.Float(width)
            self._arrow_width = width

    @property
    def etree_element(self) -> ET.Element:
        element = ET.Element("polygon")
        root = osvg.positions.PolarShiftedPosition(
            origin=self.position,
            angle=self.angle + 180,
            distance=self.arrow_length,
        )
        upper = osvg.positions.PolarShiftedPosition(
            origin=root,
            angle=self.angle + 270,
            distance=self.arrow_width / 2,
        )
        lower = osvg.positions.PolarShiftedPosition(
            origin=root,
            angle=self.angle + 90,
            distance=self.arrow_width / 2,
        )
        element.set("points", " ".join([str(upper), str(self.position), str(lower)]))
        self._add_style(element=element)
        return element


class LightArrowMarkerParams(ArrowMarkerParams):
    """
    Keyword argument definition for LightArrowMarker class.
    """

    inner_length: float | osvg.float.Float = None


class LightArrowMarker(ArrowMarker):
    """
    Class for Arrow Markers, which are more light/sharp
    """

    def __init__(
        self,
        parent: "osvg.line.Line | osvg.polyline.Polyline",
        position: osvg.positions.Position,
        angle: float | osvg.float.Float,
        **kwargs: Unpack[LightArrowMarkerParams],
    ):
        super().__init__(parent=parent, position=position, angle=angle, **kwargs)
        self._inner_length = None
        self.inner_length = kwargs.get("inner_length", None)

    @property
    def length(self) -> osvg.float.Float:
        """
        Get length
        """
        return self.inner_length

    @property
    def inner_length(self) -> osvg.float.Float:
        """
        Get length
        """
        return self._inner_length

    @inner_length.setter
    def inner_length(self, inner_length: float | osvg.float.Float | type[None]) -> None:
        if inner_length is None:
            inner_length = self.arrow_length / 2
        if not isinstance(inner_length, osvg.float.Float):
            inner_length = osvg.float.Float(inner_length)
        self._inner_length = inner_length

    @property
    def etree_element(self) -> ET.Element:
        element = ET.Element("polygon")
        root = osvg.positions.PolarShiftedPosition(
            origin=self.position,
            angle=self.angle + 180,
            distance=self.arrow_length,
        )
        draw_root = osvg.positions.PolarShiftedPosition(
            origin=self.position,
            angle=osvg.float.Sum(self.angle, 180),
            distance=self.inner_length,
        )
        upper = osvg.positions.PolarShiftedPosition(
            origin=root,
            angle=osvg.float.Sum(self.angle, 270),
            distance=self.arrow_width / 2,
        )
        lower = osvg.positions.PolarShiftedPosition(
            origin=root,
            angle=osvg.float.Sum(self.angle, 90),
            distance=self.arrow_width / 2,
        )
        element.set(
            "points",
            " ".join([str(draw_root), str(upper), str(self.position), str(lower)]),
        )
        self._add_style(element=element)
        return element


class SkinnyArrowMarker(ArrowMarker):
    """
    Class for Arrow Markers, which are only two lines.
    """

    @property
    def length(self) -> osvg.float.Float:
        """
        Get length
        """
        return osvg.float.Float(0)

    @property
    def etree_element(self) -> ET.Element:
        element = ET.Element("polyline")
        root = osvg.positions.PolarShiftedPosition(
            origin=self.position,
            angle=self.angle + 180,
            distance=self.arrow_length,
        )
        upper = osvg.positions.PolarShiftedPosition(
            origin=root,
            angle=self.angle + 270,
            distance=self.arrow_width / 2,
        )
        lower = osvg.positions.PolarShiftedPosition(
            origin=root,
            angle=self.angle + 90,
            distance=self.arrow_width / 2,
        )
        element.set("points", " ".join([str(upper), str(self.position), str(lower)]))
        self._add_color_style(element=element)
        return element



"""
Module to define base Element classes.
"""

import xml.etree.ElementTree as ET
from typing import TypedDict
from typing_extensions import Unpack
import osvg.float
import osvg.float_math
import osvg.helper
import osvg.positions
import osvg.style


SVG_STROKE_WIDTH_DEFAULT = 1


class SVGElementParams(TypedDict):
    """
    Keyword argument definition for SVGElement class.
    """

    name: str = None
    parent: "SVGElement" = None
    position: osvg.positions.Position = None
    style: osvg.style.Style = None
    layer: int = 0
    rotation: "float | osvg.float.Float" = 0
    hyperlink: str = None


class SVGElement:
    """
    Parent class for Element and Group/SVG classes.
    """

    # pylint: disable=too-many-instance-attributes

    xml_tag = ""
    layer = 0
    position = osvg.positions.PositionProperty()
    element_rotation = osvg.float.FloatProperty()
    parent_rotation = osvg.float.FloatProperty()

    def __init__(self, **kwargs: Unpack[SVGElementParams]) -> None:
        self.name = kwargs.get("name", None)
        self.parent_obj = None
        self.childs = []
        parent = kwargs.get("parent", None)
        self.parent = parent
        self.element_rotation = kwargs.get("rotation", 0)
        self.rotation = osvg.float.Sum(self.element_rotation, self.parent_rotation)
        position = kwargs.get("position", None)
        if position is None and parent is not None:
            self.position = parent.position
        else:
            self.position = position
        self.style_obj = None
        self.style = kwargs.get("style", None)
        self.layer = kwargs.get("layer", 0)
        self.connectors = {}
        self.hyperlink = kwargs.get("hyperlink", None)

    @property
    def parent(self):
        """
        Get parent object
        """
        return self.parent_obj

    @parent.setter
    def parent(self, parent: "SVGElement"):
        """
        Set and validate parent
        """
        # Check parent type
        if parent is not None and not issubclass(type(parent), SVGElement):
            raise TypeError("Parent type not allowed")
        # Check ancestor loop
        if parent is not None and (self == parent or self in parent.ancestors):
            raise RuntimeError("Parent loop detected")
        # Derefer/Refer parent
        if self.parent_obj:
            self.parent_obj.remove_child(self)
        if parent is None:
            self.parent_rotation = 0
        else:
            parent.add_child(self)
            self.parent_obj = parent
            self.parent_rotation = parent.rotation

    @property
    def ancestors(self) -> list["SVGElement"]:
        """
        Get all ancestors recursively
        """
        if self.parent_obj is None:
            return []
        return self.parent_obj.ancestors + [self.parent_obj]

    @staticmethod
    def verify_other(other: "SVGElement", class_check: bool = True):
        """
        Verify if other's class is a subclass to IOR
        """
        if class_check and not issubclass(type(other), SVGElement):
            raise TypeError("Other's class is not a subclass to ElementBase")

    def add_child(self, other: "SVGElement") -> None:
        """
        Add other IOR object as a child
        """
        self.verify_other(other)
        self.childs.append(other)

    def remove_child(self, other: "SVGElement") -> None:
        """
        Remove other IOR object from childs
        """
        self.childs.remove(other)

    @property
    def descendants(self) -> dict:
        """
        Return a dictionary with all childs recursively
        if child has a name
        """
        descendants = {}
        for o in self.childs:
            if o.name is not None:
                descendants[o.name] = o
            descendants.update(o.descendants)
        return descendants

    @property
    def style(self) -> osvg.style.Style:
        """
        Get style of this element.
        """
        return self.style_obj

    @style.setter
    def style(self, style: osvg.style.Style):
        """
        Set style for this element.
        """
        if style is None:
            style = osvg.style.Style()
        elif not isinstance(style, osvg.style.Style):
            raise TypeError("Style must be a style object")
        self.style_obj = style

    @property
    def _plain_etree_element(self) -> ET.Element:
        """
        Create ET for this object without wrappers
        """
        # Create the ET Element
        element = ET.Element(self.xml_tag)
        if self.name is not None:
            element.set("id", self.name)
        if self.parent:
            style_str = str(self.parent.style.diff(self.style))
        else:
            style_str = str(self.style)
        if style_str:
            element.set("style", style_str)
        # Get the child ET Element list
        sub_elements = osvg.helper.WeightedList()
        for obj in self.childs:
            sub_ets = obj.etree_element
            if isinstance(sub_ets, list):
                for et in sub_ets:
                    sub_elements.append(et, obj.layer)
            else:
                sub_elements.append(sub_ets, obj.layer)
        for e in sub_elements:
            element.append(e)
        return element

    @property
    def etree_element(self) -> ET.Element:
        """
        Create ET element for this object
        """
        # If element has a hyperlink, this needs to be wrapped.
        if self.hyperlink is not None:
            element = ET.Element("a")
            element.set("href", self.hyperlink)
            element.append(self._plain_etree_element)
            return element
        return self._plain_etree_element

    def add_connector(
        self, position, name: str, respect_rotation=False
    ) -> osvg.positions.Position:
        """
        Add a position to the connector dictionary.
        Self must be within position's reference path.
        Return the added connector position
        """
        if name in self.connectors:
            raise ValueError(f"Connector with name {name} already exists")
        if not isinstance(position, osvg.positions.Position):
            raise TypeError("position must be a Position object")
        if respect_rotation:
            position = osvg.positions.RotatedPosition(
                origin=position,
                center=self.position,
                angle=self.rotation,
            )
        self.connectors[name] = position
        return position

    def add_hyperlink(self, hyperlink: str) -> None:
        """
        Add an hyperlink to this element
        """
        self.hyperlink = hyperlink

    @property
    def stroke_width(self) -> float:
        """
        Get stroke width style attribute of Element or upstream value
        """
        if self.style.stroke_width is not None:
            return self.style.stroke_width
        if not self.parent:
            return SVG_STROKE_WIDTH_DEFAULT
        return self.parent.stroke_width

    def _inherited_style(self, style: osvg.style.Style):
        """
        Adjust given style object with inherited style attributes
        """
        if self.parent:
            style = self.parent._inherited_style(style=style)
        style.merge(self.style)
        return style

    @property
    def inherited_style(self) -> osvg.style.Style:
        """
        Return style object with inherited merged with element's style attributes
        """
        return self._inherited_style(style=osvg.style.Style())

    def add_element_rotation(self, element: ET.ElementTree) -> None:
        """
        Add transform tag with rotation if required
        """
        if float(self.rotation) != 0:
            element.set(
                "transform",
                f"rotate({str(self.rotation)} {str(self.position.x)} {str(self.position.y)})",
            )


class Height(osvg.float.Mul):
    """
    Relative value to the heigth of an element.
    """

    def __init__(
        self,
        element: SVGElement,
        factor: osvg.float.Float | float = 1,
    ) -> None:
        if not hasattr(element, "height_reference"):
            raise TypeError(
                "Can only reference to an element with a height_reference attribute"
            )
        super().__init__(element.height_reference, factor)


class Width(osvg.float.Mul):
    """
    Relative value to the width of an area.
    """

    def __init__(
        self,
        element: SVGElement,
        factor: float | osvg.float.Float = 1,
    ) -> None:
        if not hasattr(element, "width_reference"):
            raise TypeError(
                "Can only reference to an element with a width_reference attribute"
            )
        super().__init__(element.width_reference, factor)



"""
Module to define PositionalElement classes.
"""

import xml.etree.ElementTree as ET
import xml.dom.minidom
import osvg.elements.group
import osvg.elements.rectanglebase
import osvg.style
import osvg.float


class ViewBox:
    """
    Representation of a viewBox value tuple
    """

    # pylint: disable=too-few-public-methods
    min_x = osvg.float.FloatProperty()
    min_y = osvg.float.FloatProperty()
    width = osvg.float.FloatProperty()
    height = osvg.float.FloatProperty()

    def __init__(
        self,
        width: int | float,
        height: int | float,
        min_x: int | float = 0,
        min_y: int | float = 0,
    ) -> None:
        self.min_x = min_x
        self.min_y = min_y
        self.width = width
        self.height = height

    def __str__(self) -> str:
        return f"{self.min_x} {self.min_y} {self.width} {self.height}"


class SVG(osvg.elements.rectanglebase.RectangleBase, osvg.elements.group.Group):
    """
    This is the starting point to place elements within.
    """

    xml_tag = "svg"

    def __init__(
        self,
        width: float | osvg.float.Float,
        height: float | osvg.float.Float,
        viewbox: (
            ViewBox | tuple[int | float, int | float, int | float, int | float]
        ) = None,
        style: osvg.style.Style = None,
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            style=style,
        )
        self.viewbox = None
        if viewbox is not None:
            self.set_viewbox(viewbox=viewbox)

    @property
    def etree_element(self) -> ET.Element:
        """
        Create ET element for this object
        """
        element = super().etree_element
        element.set("width", str(self.width))
        element.set("height", str(self.height))
        if self.viewbox:
            element.set("viewBox", str(self.viewbox))
        element.set("xmlns", "http://www.w3.org/2000/svg")
        return element

    @property
    def xml_string(self):
        """
        Return a pretty string with the XML code
        """
        return xml.dom.minidom.parseString(
            ET.tostring(self.etree_element)
        ).toprettyxml()

    def set_viewbox(
        self, viewbox: tuple[int | float, int | float, int | float, int | float]
    ) -> None:
        """
        Verify and set viewbox parameters.
        """
        if isinstance(viewbox, ViewBox):
            self.viewbox = viewbox
        elif not isinstance(viewbox, tuple) or len(viewbox) != 4:
            raise TypeError("viewbox must be an tuple of four values")
        else:
            self.viewbox = ViewBox(
                min_x=viewbox[0],
                min_y=viewbox[1],
                width=viewbox[2],
                height=viewbox[3],
            )



"""
Module to define Line in SVG.
"""

import xml.etree.ElementTree as ET
from collections.abc import Iterable
from typing import TypedDict
from typing_extensions import Unpack
import osvg.elements.elementbase
import osvg.elements.group
import osvg.style
import osvg.elements.line_marker
from osvg.positions import Position


class PolylineParams(TypedDict):
    """
    Keyword argument definition for Line class.
    """

    parent: osvg.elements.group.Group
    positions: Iterable[osvg.positions.Position | Iterable[float | int]]
    name: str = None
    style: osvg.style.Style = None
    layer: int = 0


class Polyline(osvg.elements.elementbase.SVGElement):
    """
    A Line in SVG over at least two positions.
    """

    xml_tag = "polyline"

    def __init__(self, **kwargs: Unpack[PolylineParams]) -> None:
        super().__init__(**kwargs)
        self.positions = []
        for p in kwargs["positions"]:
            if issubclass(type(p), Position):
                self.positions.append(p)
            else:
                self.positions.append(Position(x=p[0], y=p[1]))
        if len(self.positions) < 2:
            raise ValueError(f"{self.__class__} needs at least two positions")
        self.start_marker = None
        self.end_marker = None

    @property
    def etree_element(self) -> ET.Element:
        # pylint: disable=f-string-without-interpolation
        element = super().etree_element
        # Adjust positions according to start/end marker lenghts
        positions = self.positions[:]
        pop_first = False
        if self.start_marker and float(self.start_marker.length) > 0:
            distance = osvg.positions.Distance(a=self.start, b=self.positions[1])
            if float(distance) > float(self.start_marker.length):
                positions[0] = osvg.positions.PolarShiftedPosition(
                    origin=self.start,
                    angle=osvg.positions.AngleDegree(a=self.start, b=self.positions[1]),
                    distance=self.start_marker.length,
                )
            else:
                pop_first = True
        pop_last = False
        if self.end_marker and float(self.end_marker.length) > 0:
            distance = osvg.positions.Distance(a=self.positions[-2], b=self.end)
            if float(distance) > float(self.end_marker.length):
                positions[-1] = osvg.positions.PolarShiftedPosition(
                    origin=self.end,
                    angle=osvg.positions.AngleDegree(a=self.end, b=self.positions[-2]),
                    distance=self.end_marker.length,
                )
            else:
                pop_last = True
        if pop_first:
            positions = positions[1:]
        if pop_last:
            positions = positions[:-1]

        # Add OSVG element infos
        element.set("points", " ".join([str(p) for p in positions]))
        if self.start_marker or self.end_marker:
            element = [element]
            if self.start_marker:
                element.append(self.start_marker.etree_element)
            if self.end_marker:
                element.append(self.end_marker.etree_element)
        return element

    @property
    def start(self) -> osvg.positions.Position:
        """
        Return first position
        """
        return self.positions[0]

    @property
    def end(self) -> osvg.positions.Position:
        """
        Return last position
        """
        return self.positions[-1]

    def add_start_marker(
        self, marker_class: osvg.elements.line_marker.Marker, **kwargs
    ) -> None:
        """
        Add a start marker to the line
        """
        angle = osvg.positions.AngleDegree(a=self.positions[1], b=self.start)
        self.start_marker = marker_class(
            parent=self, position=self.start, angle=angle, **kwargs
        )

    def add_end_marker(
        self, marker_class: osvg.elements.line_marker.Marker, **kwargs
    ) -> None:
        """
        Add a end marker to the line
        """
        angle = osvg.positions.AngleDegree(a=self.positions[-2], b=self.end)
        self.end_marker = marker_class(
            parent=self, position=self.end, angle=angle, **kwargs
        )
